#include <iostream>
#include <string>
#include <set>
#include <map>
#include <stack>

#include "linkdb.h"
#include "sitemap.h"

class collectURLs : public urlVisitor
{
	public:
		collectURLs( siteMap& sitemap, linkDB& db ) : count( 0 ), sitemap( sitemap ), db( db ) {}
			
		virtual void visit( const md5sum& hash, const string& url, const string& title )
		{
			md5sum siteId;
			
			// If this URL has not been assigned to a site, go build the site information
			if ( ! sitemap.getId( hash, siteId ) )
			{
				// Otherwise, we need to collect all the links in this site
				map<md5sum, vector<md5sum> > urlLinks;
				set<md5sum> searchSet;
				
				searchSet.insert( hash );
	
				// Iterate over all the URLs in the site, adding new ones as we discover them
				while ( ! searchSet.empty() )
				{
					set<md5sum>::iterator top = searchSet.begin();
					md5sum urlId = *top;
					searchSet.erase( top );
					
					// It is impossible to have seen this URL already
					assert( urlLinks.find( urlId ) == urlLinks.end() );
					
					// Get the md5sums of all the links associated with this URL
					vector<md5sum> links = db.getLinks( urlId );
					assert( links.size() > 0 );
					
					// Store the link information for the second pass
					urlLinks.insert( pair<md5sum, vector<md5sum> >( urlId, links ) );
					
					for ( vector<md5sum>::const_iterator i = links.begin(); i != links.end(); ++ i )
					{
						// If we havn't seen this URL, add it to the searchStack
						if ( urlLinks.find( *i ) == urlLinks.end() )
							searchSet.insert( *i );
					}
				}
				
				// At this point, we have a collection of all the URLs in this site
				// and we need to build structure
				
				// Find the URL with the lowest URL string in sorted order
				string leastURL = url;
				siteId = hash;
				for ( map<md5sum, vector<md5sum> >::const_iterator i = urlLinks.begin(); i != urlLinks.end(); ++ i )
				{
					string urlString = db.getURLString( i->first );
					assert( urlString != "" );
					
					if ( urlString < leastURL )
					{
						leastURL = urlString;
						siteId = i->first;
					}
				}
				
				// Now recursively build the tree
				map<md5sum, vector<md5sum> >::iterator leastURLIt = urlLinks.find( siteId );
				assert( leastURLIt != urlLinks.end() );
				vector<md5sum> leastURLLinks = leastURLIt->second;
				urlLinks.erase( leastURLIt );
				
				Tree siteStructure( siteId );
				nodeCount = 1;
				addLinkedURLs( &siteStructure, leastURLLinks, urlLinks );
				
				assert( urlLinks.size() == 0 );
				
				// Save the site structure
				sitemap.serializeSite( siteId, siteStructure );
			}
			
			++ count;
			if ( count % 100 == 0 )
			{
				cerr << ".";
				if ( count % 1000 == 0 )
					cerr << count << endl;
			}
		}
		
		// The destructor is virtual because G++ told me so. Isn't it nice?
		virtual ~collectURLs()
		{
			cerr << endl << count << " URLs processed" << endl;
		}
	
	private:
		// TODO: This recursive function will hog RAM and is probably as slow as garbage. Oh well.
		void addLinkedURLs( TreeNode* node, vector<md5sum> links, map<md5sum, vector<md5sum> >& remainingURLs )
		{
			assert( remainingURLs.find( node->getValue() ) == remainingURLs.end() );
			
			vector<vector<md5sum> > childLinks;
			
			for ( vector<md5sum>::const_iterator i = links.begin(); i != links.end(); ++ i )
			{
				// If the URL has not been parented by another site, we'll take it!
				map<md5sum, vector<md5sum> >::iterator it = remainingURLs.find( *i );
				if ( it != remainingURLs.end() )
				{
					node->appendChild( *i );
					++ nodeCount;
					childLinks.push_back( it->second );
					remainingURLs.erase( it );
				}
			}
			
			assert( node->size() == childLinks.size() );
			
			// Now call again for all the children
			int j = 0;
			for ( vector<TreeNode*>::iterator i = node->begin(); i != node->end(); ++ i, ++ j )
			{
				addLinkedURLs( *i, childLinks[j], remainingURLs );
			}
		}
		
		unsigned long nodeCount;
		
		unsigned long count;
		siteMap& sitemap;
		linkDB& db;
};

void usage()
{
	cout << "./gen-sitemap [options]" << endl
		<< "Options:" << endl
		<< "\t-h -?\tHelp: Show this usage message" << endl
		<< "\t-d <dir>\tDirectory: Use the databases in <dir>" << endl;
}

int main( int argc, char** argv )
{
	string outputPath = "/tmp/bidir";
	
	// -? : help
	// -h : help
	// -d : Use database in directory ...
	char options[] = "?hd:";
	
	char opt = ' ';
	while ( (opt = getopt( argc, argv, options )) != -1 )
	{
		switch ( opt )
		{
			// Usage
			case 'h':
			case '?':
				usage();
				return 0;
				break;
			
			case 'd':
				outputPath = optarg;
				break;				
		}
	}
	
	linkDB db( outputPath, true );
	siteMap sitemap( outputPath );

	// Collect all the URLs into sites
	collectURLs visitor( sitemap, db );
	db.visitURLs( visitor );
	
	return 0;
}
