#include "linkdb.h"

#include "basictypes.h"

using std::string;

const char linkDB::DATABASE_HASH_TO_URL[] = "google-urls.db";
const char linkDB::DATABASE_LINKS_TO[] = "google-links-forward.db";
//const char* linkDB::DATABASE_LINKED_FROM = "/tmp/google-links-backward.db";

string linkDB::getURLString( const md5sum& hash )
{
	string returnValue;
	string title;
	
	getURLTitlePair( hash, returnValue, title );
	return returnValue;
}

void linkDB::unpackURLTitlePair( const Dbt& data, std::string& url, std::string& title ) const
{
	char* urlBuffer = (char*)data.get_data() + sizeof( uint32 );
	uint32 urlLength = *((uint32*)data.get_data());
	char* titleBuffer = urlBuffer + urlLength;
	uint32 titleLength = data.get_size() - sizeof( uint32 ) - urlLength;
	
	// "Return" values
	url = string(  urlBuffer, urlLength );
	title = string( titleBuffer, titleLength );
}

void linkDB::getURLTitlePair( const md5sum& hash, std::string& url, std::string& title )
{
	// Key is the hash
	Dbt key;
	key.set_data( (void*) hash.getDigest() );
	key.set_size( hash.size() );
	
	// Data is the URL/Title pair:
	// 4 bytes: u_int_32 url.size()
	// x bytes: url.data()
	// x bytes: title.data()
	Dbt data;

	// Get the original value in the database
	int result = hashToURL.get( NULL, &key, &data, 0 );
	if ( result == DB_NOTFOUND )
	{
		url = string();
		title = string();
		return;
	}
	else if ( result != 0 )
	{
		// TODO: This should never happen
		cerr << "Some weird error while fetching url string\n";
		throw "TODO: Throw a real exception here";
	}
	
	unpackURLTitlePair( data, url, title );
}


void linkDB::putURL( std::string url, std::string title = "" )
{
	assert( url != "" );
	
	md5sum hash( url );
	
	// Key is the hash
	Dbt key;
	key.set_data( (void*) hash.getDigest() );
	key.set_size( hash.size() );
	
	// Data is the size + URL + title. Allocate the required space
	int bufferSize = sizeof( uint32 ) + url.size() + title.size();
	char* buffer = new char[ bufferSize ];
	// First 4 bytes: The size of the url
	*((uint32*)buffer) = url.size();
	
	// Copy the URL and Title to the buffer
	url.copy( buffer + sizeof( uint32 ), url.size() );
	title.copy( buffer + sizeof( uint32 ) + url.size(), title.size() );
	
	Dbt data;
	data.set_data( (void*) buffer );
	data.set_size( bufferSize );
	
	// Try to store the mapping
	int retVal = hashToURL.put( NULL, &key, &data, DB_NOOVERWRITE );
	// If we already stored the value, make sure it matches
	// TODO: This could be optimized away. How useful are these checks?
	if ( retVal == DB_KEYEXIST )
	{
		// Get the original value in the database
		string originalURL;
		string originalTitle;
		getURLTitlePair( hash, originalURL, originalTitle );
		
		// This should be impossible (we know that the string is in the database)
		assert( originalURL != "" );
		// This is *very* improbable: md5 has a *very* low chance of collision
		assert( originalURL == url );

		// If the original title is empty, store the new data
		if ( originalTitle == "" && title != "" )
		{
			//cerr << "Warning: updating empty title data" << endl;
			if ( hashToURL.put( NULL, &key, &data, 0 ) )
			{
				cerr << "Some error occured storing the URL/Title pair" << endl;
				delete buffer;
				throw "TODO: Throw a real exception here";
			}
		}
	}
	else if ( retVal )
	{
		cerr << "Some error occured storing the URL/Title pair" << endl;
		delete buffer;
		throw "TODO: Throw a real exception here";
	}
	
	delete buffer;
}


void linkDB::link( string src, string dest )
{
	md5sum srcHash( src );
	md5sum destHash( dest );
	
	// Store the md5 -> url mappings
	putURL( src );
	putURL( dest );
	
	// Index the links
	link( srcHash, destHash );
}

void linkDB::link( const md5sum& src, const md5sum& dest )
{
	// Index the links
	// Only add the linkedFrom entry if the linksTo entry succeeded.
	// The linksTo entry can fail if this is a duplicate
	if ( addURL( linksTo, src, dest ) )
	{
		/*
		if ( ! addURL( linkedFrom, destHash, srcHash ) )
		{
			cerr << "Adding the linkedFrom entry failed\n";
			throw "TODO: Throw a real exception here";
		}
		*/
	}
}


bool linkDB::isLinked( const md5sum& src, const md5sum& dest )
{
	// Key is the source hash
	Dbt key;
	key.set_data( (void*) src.getDigest() );
	key.set_size( src.size() );
	
	// Data is the destination hash
	Dbt data;
	data.set_data( (void*) dest.getDigest() );
	data.set_size( dest.size() );

	int result = linksTo.get( NULL, &key, &data, DB_GET_BOTH );
	
	assert( result == 0 || result == DB_NOTFOUND );
	
	return result == 0;
}

bool linkDB::exists( const std::string& url )
{
	md5sum hash( url );
	// The URL exists in the database if we map the hash -> url and get something
	return ( getURLString( hash ).length() > 0 );
}

std::vector<md5sum> linkDB::getLinks( const std::string& url )
{
	return this->getLinks( md5sum( url ) );
}

std::vector<md5sum> linkDB::getLinks( const md5sum& hash )
{
	Dbt key;
	key.set_data( (void*) hash.getDigest() );
	key.set_size( hash.size() );

	Dbc* cursor;
	linksTo.cursor( NULL, &cursor, 0 );
	
	std::vector<md5sum> links;
	
	Dbt data;
	int retVal = cursor->get( &key, &data, DB_SET );
	
	while( retVal == 0 )
	{
		md5sum value( (unsigned char *) data.get_data() );
		links.push_back( value );
		retVal = cursor->get( &key, &data, DB_NEXT_DUP );
	}
	
	// cursor->get should have returned DB_NOTFOUND to get here
	assert( retVal == DB_NOTFOUND );
		
	cursor->close();
	
	return links;
}

bool linkDB::addURL( Db& database, const md5sum& hash, const md5sum& value )
{
	// Key is the hash
	Dbt key;
	key.set_data( (void*) hash.getDigest() );
	key.set_size( hash.size() );
	
	// Data is another hash
	Dbt data;
	data.set_data( (void*) value.getDigest() );
	data.set_size( value.size() );
	
	// store it back in the database
	int result = database.put( NULL, &key, &data, DB_NODUPDATA  );
	if ( DB_KEYEXIST == result )
	{
		// This occurs *a lot* (eg. any web page with two links to the same page on it)
		//cerr << "Warning: Duplicate link found\n";
		return false;
	}
	else if ( result != 0 )
	{
		cerr << "An error occurred adding a URL to the database." << endl;
		throw "TODO: Throw a real exception here";
	}
	
	return true;
}

void linkDB::visitURLs( urlVisitor& visitor )
{
	Dbt key;
	Dbt data;
		
	Dbc* cursor;
	hashToURL.cursor( NULL, &cursor, 0 );
	
	int retVal = cursor->get( &key, &data, DB_FIRST );	
	while( retVal != DB_NOTFOUND )
	{
		string url;
		string title;
		unpackURLTitlePair( data, url, title );
		
		visitor.visit( md5sum( (unsigned char *) key.get_data() ), url, title );
		
		retVal = cursor->get( &key, &data, DB_NEXT );
	}
	
	cursor->close();
}

void linkDB::visitLinks( rawLinkVisitor& visitor )
{
	Dbt key;
	Dbt data;
		
	Dbc* cursor;
	linksTo.cursor( NULL, &cursor, 0 );
	
	int retVal = cursor->get( &key, &data, DB_FIRST );	
	while( retVal != DB_NOTFOUND )
	{
		// Retrieve the hashes and look them up
		md5sum srcHash( (unsigned char*) key.get_data() );
		md5sum destHash( (unsigned char*) data.get_data() );
		
		// Visit the link
		visitor.visit( this, srcHash, destHash );
				
		// Move to the next link
		retVal = cursor->get( &key, &data, DB_NEXT );
	}
	
	cursor->close();
}

