#ifndef _LINKDB_H
#define _LINKDB_H

#include <vector>
#include <string>
#include <sstream>
#include <iomanip>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <db_cxx.h>

#include "md5sum.h"

/** Visitor interface for the visitURLs method. Visits each URL in the database. */
class urlVisitor
{
public:
	virtual void visit( const md5sum& hash, const string& url, const string& title ) = 0;
};

class linkDB;

/** Visitor interface for the visitLinks method. Visits each link in the database. */
class rawLinkVisitor
{
public:
	virtual void visit( linkDB* db, const md5sum& source, const md5sum& destination ) = 0;
};

class linkDB
{
public:
	/** Default constructor: Creates/opens the link database in the specified output path. 
	* @param specifiedOutputPath The path to create/read the databases. Defaults to "/tmp".
	* @param readOnly Specifies if the databases should be opened read only. Read only databases can have better performance. Defaults to false. */
	linkDB( const std::string& specifiedOutputPath = "", bool readOnly = false ) :
		hashToURL( NULL, 0 ),
		linksTo( NULL, 0 ),
		//linkedFrom( NULL, 0 )
		outputPath( specifiedOutputPath )
	{
		// Default to "/tmp"
		if ( outputPath.empty() )
			outputPath = "/tmp";
		
		// Verify that the path exists and is a directory
		struct stat fileInfo;
		if ( stat( outputPath.c_str(), &fileInfo ) )
		{
			// TODO: Throw a real exception here
			throw "Invalid output path (stat failed). Maybe it does not exist?";
		}
		
		// If it isn't a directory, throw an exception
		if ( ! S_ISDIR( fileInfo.st_mode ) )
		{
			//TODO: Throw a real exception here
			throw "Invalid output path (is not a directory)";
		}

		// Add a trailing slash if it was not specified
		if ( '/' != *outputPath.rbegin() )
		{
			outputPath.append( "/" );
		}
		
		// Make sure that the links to and linked from databases allow duplicate keys
		//cerr << "setting flags" << endl;
		linksTo.set_flags( DB_DUPSORT );
		//linkedFrom.set_flags( DB_DUPSORT );
		//cerr << "created" << endl;
		
		u_int32_t flags = DB_CREATE;
		if ( readOnly )
			flags = DB_RDONLY;
		
		// Open the three databases
		hashToURL.open( (outputPath + DATABASE_HASH_TO_URL).c_str(), NULL, DB_HASH, flags, 0 );
		linksTo.open( (outputPath + DATABASE_LINKS_TO).c_str(), NULL, DB_HASH, flags, 0 );
		//linkedFrom.open( DATABASE_LINKED_FROM, NULL, DB_HASH, DB_CREATE, 0 );
		
		
	}
	
	/** Closes the open database handles. */
	~linkDB()
	{
		hashToURL.close( 0 );
		linksTo.close( 0 );
		//linkedFrom.close( 0 );
	}
	
	/** Stores a URL and Title in the md5 -> url database. */
	void putURL( std::string url, std::string title = "" );
	
	/** Records a link from the source URL to the destination URL.
	* @param src The URL or MD5 of the page that the link is on.
	* @param dest The URL or MD5 of the page that the link leads to. */
	void link( std::string src, std::string dest );
	void link( const md5sum& src, const md5sum& dest );
	
	/** Tests if the specified link exists in the database. */
	//bool isLinked( std::string srcUrl, std::string destUrl );
	bool isLinked( const md5sum& src, const md5sum& dest );
	
	/** Tests if the specified URL exists in the database. */
	bool exists( const std::string& url );
	
	/** Returns a vector of hashes that the specified URL links to. */
	std::vector<md5sum> getLinks( const std::string& url );
	std::vector<md5sum> getLinks( const md5sum& hash );
	
	/** Iterate through every single document in the database. */
	void visitURLs( urlVisitor& visitor );
	
	/** Iterate through all the links in the database. */
	void visitLinks( rawLinkVisitor& visitor );
	
	/** Find the URL string for a given MD5 hash.
	* @param hash The MD5 sum of the URL to be returned.
	* @returns The URL string for the hash, or an empty string if it does not exist in the database. */
	std::string getURLString( const md5sum& hash );
	
	/** Returns the URL/Title pair for a given MD5 hash.
	* @param hash The MD5 sum of the URL to be returned.
	* @param url The variable to return the URL in.
	* @param title The variable to return the Title in.
	*/
	void getURLTitlePair( const md5sum& hash, std::string& url, std::string& title );
	
private:
	/** Adds a URL to the collection with the specified key in the specified database. */
	bool addURL( Db& database, const md5sum& key, const md5sum& value );

	/** Unpacks and returns the URL/Title pair for a given data buffer.
	* @param data The data buffer from the Berkeley DB database containing the URL and Title.
	* @param url The variable to return the URL in.
	* @param title The variable to return the Title in.
	*/
	void unpackURLTitlePair( const Dbt& data, std::string& url, std::string& title ) const;

	static const char DATABASE_HASH_TO_URL[];
	static const char DATABASE_LINKS_TO[];
	//static const char* DATABASE_LINKED_FROM;

	Db hashToURL;
	Db linksTo;
	//Db linkedFrom;	

	std::string outputPath;
};


/** Visitor interface for the visitLinks method. Visits each link in the database. */
class linkVisitor : public rawLinkVisitor
{
public:
	virtual void visit( linkDB* db, const md5sum& srcHash, const md5sum& destHash )
	{
		string src = db->getURLString( srcHash );
		string dest = db->getURLString( destHash );
		
		// TODO: This should never happen. It means the database is incorrect
		if ( src == "" || dest == "" )
		{
			cerr << "Found links with no associated URL strings." << endl;
			throw "TODO: Throw a real exception here";			
		}		
		
		// Call the callback
		this->visit( src, dest );
	}
	
	virtual void visit( const string& source, const string& destination ) = 0;
};

#endif
