//***************************************************************************
// This source code is copyrighted 2002 by Google Inc.  All rights
// reserved.  You are given a limited license to use this source code for
// purposes of participating in the Google programming contest.  If you
// choose to use or distribute the source code for any other purpose, you
// must either (1) first obtain written approval from Google, or (2)
// prominently display the foregoing copyright notice and the following
// warranty and liability disclaimer on each copy used or distributed.
// 
// The source code and repository (the "Software") is provided "AS IS",
// with no warranty, express or implied, including but not limited to the
// implied warranties of merchantability and fitness for a particular
// use.  In no event shall Google Inc. be liable for any damages, direct
// or indirect, even if advised of the possibility of such damages.
//***************************************************************************


// Print all the URLs that a file links to

#include <string>
#include <algorithm>

#include <stdio.h>
#include "handler-parser.h"        // where parsehandlers are defined

#include "parsehandler-preparse.h"

#include "linkdb.h"

#include "url_string.hh"

using std::string;

class BuildLinkDBParseHandler : public ParseHandler {
public:
	BuildLinkDBParseHandler() : count( 0 ) {}
	
	virtual void NewDocument(const Document* doc)
	{
		docURL = string( doc->url() );
		baseURL = docURL;
		
		inHead = false;
		savedURL = false;
		docTitle = "";

		/* REPLACED BY LIBURL
		// Break it up so we can create turn relative URLs into "fully qualified" URLs
		breakURLIntoParts( docURL );
		*/
		/*
		string bug( "http://www.alumcareer.ufl.edu/AlumniLogin.html" );
		string url = docURL.get_url();
		url.c_str();
		bug.c_str();
		if ( url == bug )
		{
			cout << "Found it" << endl;
		}
		*/
	}

	// Any text while in the "header" will get added to the "title". This is inaccurate, but is
	// the best we can get from Google's parser
	virtual void AddHead( const char* fields, int fieldlen )
	{
		inHead = true;
	}
	
	// Any text term found while in the header is added to the title
	virtual void AddTerm(const char* term, int termlen, int face, int size)
	{
		if ( inHead )
		{
			docTitle.append( term, termlen );
		}
	}
	
	// Any punctuation term found while in the header is added to the title
	virtual void AddPunctuation(const char * text, int textlen, int face, int size)
	{
		if ( inHead )
		{
			docTitle.append( text, textlen );
		}
	}
	
	// Any whitespace term found while in the header is added to the title as a space ' '
	virtual void WhitespaceEndedTerm()
	{
		if ( inHead )
		{
			docTitle.append( " " );
		}
	}
	
	// The header is over, save the document URL and title
	virtual void AddHeadDone()
	{
		inHead = false;
		if ( ! docTitle.empty() )
		{
			// Trim trailing space off of the title. There should only be one, but we might as well make sure
			while ( *docTitle.rbegin() == ' ' )
				docTitle.resize( docTitle.size() - 1 );
		}
		
		db.putURL( docURL.get_url(), docTitle );
		savedURL = true;
	}
	
	// If there was no header, we need to verify that we have saved this URL in the database
	virtual void EndDocument(const Document* doc)
	{
		if ( ! savedURL )
		{
			db.putURL( docURL.get_url(), docTitle );
			savedURL = true;
		}
		
		++ count;
		if ( count % 100 == 0 )
		{
			cerr << ".";
			if ( count % 1000 == 0 ) cerr << count << endl;
		}
	}
	
	virtual void AddBaseURL(const char* baseurl, int baseurllen)
	{
		// TODO: Does Google make the baseurl fully qualified for us? It seems like it might
		// The base URL takes precedence over the document URL to define where relative
		// links go. Therefore, we need to use the URL parser to create a "fully qualified" URL
		// then we need to break it into parts just like the document URL
		// TODO: Read the standard to make sure this is true.
		/* REPLACED BY LIBURL
		string baseURL = makeURLFullyQualified( string( baseurl, baseurllen ) );
		breakURLIntoParts( baseURL );
		*/
		string url = string( baseurl, baseurllen );
		baseURL = url_string( url, baseURL );
		//cout << docURL.get_url() << " + base: " << url << " = base url: " << baseURL.get_url() << endl;
	}
	
	virtual void AddAnchor(const char* href, int hreflen)
	{
		// Get the fully qualified URL, verify it, and then store it
		/* REPLACED BY LIBURL
		string url = makeURLFullyQualified( string( href, hreflen ) );
		*/
		string originalHref = string( href, hreflen );
		
		// Chop off any '#' fragments
		// TODO: What about query strings? ('?')
		string::size_type pos = originalHref.find( '#' );
		if ( string::npos != pos )
		{
			originalHref.resize( pos );
		}
	
		url_string url( originalHref, baseURL );
		const string& absoluteURL = url.get_url();		// URL filter:
		// Protocol must be http, https, or ftp
		const string& scheme = url.get_scheme();
		if ( scheme == "http" || scheme == "https" || scheme == "ftp" )
		{
			bool urlIsValid = true;
			for ( string::const_iterator i = absoluteURL.begin(); i != absoluteURL.end(); ++ i )
			{
				if ( static_cast<unsigned char>( *i ) & 0x80 )
				{
					urlIsValid = false;
					break;
				}
			}
			
			if ( urlIsValid )
			{
				db.link( docURL.get_url(), absoluteURL );
			}
			else
			{
				//cerr << "Invalid URL: '" << absoluteURL << "'" << endl << "\tIn document: '" << docURL.get_url() << "'" << endl;
				return;
			}
		}
		//else if ( scheme != "mailto" && scheme != "gopher" && scheme != "file" && scheme != "javascript" && scheme != "telnet" && scheme != "news" && scheme != "about" )
		//	cerr << "Warning: ignoring link: " << absoluteURL << endl;

		/* REPLACED BY LIBURL
		// protocol (http)
		string::size_type pos = url.find( ":" );
		// Make sure the protocol is lowercase
		string proto = url.substr( 0, pos );
		transform( proto.begin(), proto.end(), proto.begin(), tolower );
		
		// URL filter:
		// ^(http:)|(https:)|(ftp:).*(?!.jpg)(?!.gif)$
		if ( proto == "http" || proto == "https" || proto == "ftp" )
		{
			db.link( docURL, url );
		}
		else if ( proto != "mailto" && proto != "gopher" && proto != "file" && proto != "javascript" && proto != "telnet" )
			cerr << "Warning: ignoring link: " << url << endl;
		*/
	}
	
private:
	// Breaks the URL into its parts and stores it in private members
	/* REPLACED BY LIBURL
	void breakURLIntoParts( const string url )
	{
		// protocol (http)
		string::size_type pos = url.find( "://" );
		// Make sure the protocol is lowercase
		protocol = url.substr( 0, pos );
		transform( protocol.begin(), protocol.end(), protocol.begin(), tolower );
		pos += 3;
		
		// server name
		string::size_type pos2 = url.find( '/', pos );
		serverName = url.substr( pos, pos2 - pos );
		
		// port (80)
		pos = serverName.find( ':', pos );
		if ( pos == string::npos )
		{
			port = "";
		}
		else
		{
			port = serverName.substr( pos + 1 );
			serverName.erase( pos, string::npos );
			if ( port == "80" ) port = "";
		}
		
		// file
		pos = url.rfind( '/' );
		file = url.substr( pos + 1 );
		
		// path (contains the trailing slash, not the leading slash
		path = url.substr( pos2 + 1, pos - pos2 );
	}
	*/
	
	// Takes the relative URL and returns a fully qualified URL, using this
	// class' private members
	/* REPLACED BY LIBURL
	string makeURLFullyQualified( string relativeURL ) const
	{
		// If there is an anchor reference (somefile.html#someReference) erase it
		string::size_type referencePos = relativeURL.find( '#' );
		if ( referencePos != string::npos )
			relativeURL.erase( referencePos );
		
		// Search for protocol
		string::size_type pos = relativeURL.find( "://" );
		if ( pos != string::npos )
		{
			// We found a protocol, then this *is* a fully qualified URL
			// Make sure the protocol is lowercase
			transform( relativeURL.begin(), relativeURL.end(), relativeURL.begin(), tolower );

			// We just need to make sure that it isn't a server without a trailing slash
			if ( relativeURL.find( '/', pos + 3 ) == string::npos )
				return relativeURL + "/";
			else
				return relativeURL;
		}
		
		// Check for mailto links (special case)
		if ( relativeURL.substr( 0, 7 ) == "mailto:" )
		{
			return relativeURL;
		}
		
		// See if the path is absolute
		if ( relativeURL[0] == '/' )
		{
			return makeURL( relativeURL, "" );
		}
		
		// Otherwise, the relativeURL replaces the file
		return makeURL( relativeURL );
	}
	*/
	// Builds a fully qualified URL out of the seperate parts, using the data in the private
	// member variables
	/* REPLACED BY LIBURL
	string makeURL( string fileParam = "", string pathParam = "", string serverNameParam = "", string portParam = "", string protocolParam = "" ) const
	{
		if ( fileParam == "" ) fileParam = file;
		if ( pathParam == "" ) pathParam = path;
		if ( serverNameParam == "" ) serverNameParam = serverName;
		if ( portParam == "" ) portParam = port;
		if ( protocolParam == "" ) protocolParam = protocol;
		
		if ( portParam != "" ) portParam += ":";
		
		return protocolParam + "://" + serverNameParam + portParam + "/" + resolvePath( pathParam + fileParam );
	}
	*/
	// Resolves any ".." references in a path
	/* REPLACED BY LIBURL
	string resolvePath( string path ) const
	{
		string::size_type findPos = 0;
		string::size_type pos = string::npos;
		while( (pos = path.find( "../", findPos )) != string::npos )
		{
			// Each time we find a "/../", we erase the "[^/]/" section before it.
			string::size_type erasePos = 0;
			if ( pos >= 2 )
			{
				// Make sure there is a "/" in front
				if ( path[pos - 1] != '/' )
				{
					findPos = pos + 3;
					continue;
				}
				
				erasePos = path.rfind( '/', pos - 2 );
				if ( erasePos == string::npos )
					erasePos = 0;
				else	// Keep the slash
					erasePos ++;
			}
			
			path.erase( erasePos, pos - erasePos + 3 );
			
			// Search for the next "/../" sequence starting at the pont where we erased stuff
			findPos = erasePos;
		}
	
		return path;
	}

	string docURL;
	string baseURL;
	string protocol;
	string port;
	string serverName;
	string path;
	string file;
	*/
	
	unsigned long int count;
	
	bool inHead;
	bool savedURL;
	string docTitle;
	
	url_string docURL;
	url_string baseURL;
	
	linkDB db;
};

ParseHandler* MakeBuildLinkDBHandler() {
  return new BuildLinkDBParseHandler;
}
