//***************************************************************************
// This source code is copyrighted 2002 by Google Inc.  All rights
// reserved.  You are given a limited license to use this source code for
// purposes of participating in the Google programming contest.  If you
// choose to use or distribute the source code for any other purpose, you
// must either (1) first obtain written approval from Google, or (2)
// prominently display the foregoing copyright notice and the following
// warranty and liability disclaimer on each copy used or distributed.
// 
// The source code and repository (the "Software") is provided "AS IS",
// with no warranty, express or implied, including but not limited to the
// implied warranties of merchantability and fitness for a particular
// use.  In no event shall Google Inc. be liable for any damages, direct
// or indirect, even if advised of the possibility of such damages.
//***************************************************************************


// Print all the URLs that a file links to

#include <string>

#include <stdio.h>
#include "handler-parser.h"        // where parsehandlers are defined

using std::string;

class CatAnchorParseHandler : public ParseHandler {
public:
	CatAnchorParseHandler() :
		// Make sure all the member variables have sane initial values
		docURL( "" ),
		baseURL( "" ),
		protocol( "" ),
		port( "" ),
		serverName( "" ),
		path( "" ),
		file( "" )
		{}
		
	virtual void NewDocument(const Document* doc)
	{
		docURL = doc->url();
		
		// Break it up so we can create turn relative URLs into "fully qualified" URLs
		breakURLIntoParts( docURL );
	}
	
	virtual void AddBaseURL(const char* baseurl, int baseurllen)
	{
		// The base URL takes precedence over the document URL to define where relative
		// links go. Therefore, we need to use the URL parser to create a "fully qualified" URL
		// then we need to break it into parts just like the document URL
		// TODO: Read the standard to make sure this is true.
		
		string baseURL = makeURLFullyQualified( string( baseurl, baseurllen ) );
		breakURLIntoParts( baseURL );
	}
	
	virtual void AddAnchor(const char* href, int hreflen)
	{
		// Get the fully qualified URL, verify it, and then store it
		string url = makeURLFullyQualified( string( href, hreflen ) );
		
		bool urlIsValid = true;
		for ( string::const_iterator i = url.begin(); i != url.end(); ++ i )
		{
			if ( static_cast<unsigned char>( *i ) & 0x80 )
			{
				urlIsValid = false;
				break;
			}
		}
		
		if ( urlIsValid )
		{
			cout << docURL << "    " << url << endl;
		}
	}
	
	
private:
	// Breaks the URL into its parts and stores it in private members
	void breakURLIntoParts( const string url )
	{
		// protocol (http)
		string::size_type pos = url.find( "://" );
		protocol = url.substr( 0, pos );
		pos += 3;
		
		// server name
		string::size_type pos2 = url.find( '/', pos );
		serverName = url.substr( pos, pos2 - pos );
		
		// port (80)
		pos = serverName.find( ':', pos );
		if ( pos == string::npos )
		{
			port = "";
		}
		else
		{
			port = serverName.substr( pos + 1 );
			serverName.erase( pos, string::npos );
			if ( port == "80" ) port = "";
		}
		
		// file
		pos = url.rfind( '/' );
		file = url.substr( pos + 1 );
		
		// path (contains the trailing slash, not the leading slash
		path = url.substr( pos2 + 1, pos - pos2 );
		
		/*
		fprintf( stderr, "%s\n", url.c_str() );
		printf( "%s\n", makeURL().c_str() );
		*/
	}
	
	// Takes the relative URL and returns a fully qualified URL, using this
	// class' private members
	string makeURLFullyQualified( string relativeURL ) const
	{
		// If there is an anchor reference (somefile.html#someReference) erase it
		string::size_type referencePos = relativeURL.find( '#' );
		if ( referencePos != string::npos )
			relativeURL.erase( referencePos );
		
		// Search for protocol
		string::size_type pos = relativeURL.find( "://" );
		if ( pos != string::npos )
		{
			// We found a protocol, then this *is* a fully qualified URL
			// We just need to make sure that it isn't a server without a trailing slash
			if ( relativeURL.find( '/', pos + 3 ) == string::npos )
				return relativeURL + "/";
			else
				return relativeURL;
		}
		
		// Check for mailto links (special case)
		if ( relativeURL.substr( 0, 7 ) == "mailto:" )
		{
			return relativeURL;
		}
		
		// See if the path is absolute
		if ( relativeURL[0] == '/' )
		{
			return makeURL( relativeURL, "" );
		}
		
		// Otherwise, the relativeURL replaces the file
		return makeURL( relativeURL );
	}
	
	// Builds a fully qualified URL out of the seperate parts, using the data in the private
	// member variables
	string makeURL( string fileParam = "", string pathParam = "", string serverNameParam = "", string portParam = "", string protocolParam = "" ) const
	{
		if ( fileParam == "" ) fileParam = file;
		if ( pathParam == "" ) pathParam = path;
		if ( serverNameParam == "" ) serverNameParam = serverName;
		if ( portParam == "" ) portParam = port;
		if ( protocolParam == "" ) protocolParam = protocol;
		
		if ( portParam != "" ) portParam += ":";
		
		return protocolParam + "://" + serverNameParam + portParam + "/" + resolvePath( pathParam + fileParam );
	}
	
	// Resolves any ".." references in a path
	string resolvePath( string path ) const
	{
		string::size_type findPos = 0;
		string::size_type pos = string::npos;
		while( (pos = path.find( "../", findPos )) != string::npos )
		{
			// Each time we find a "/../", we erase the "[^/]/" section before it.
			string::size_type erasePos = 0;
			if ( pos >= 2 )
			{
				// Make sure there is a "/" in front
				if ( path[pos - 1] != '/' )
				{
					findPos = pos + 3;
					continue;
				}
				
				erasePos = path.rfind( '/', pos - 2 );
				if ( erasePos == string::npos )
					erasePos = 0;
				else	// Keep the slash
					erasePos ++;
			}
			
			path.erase( erasePos, pos - erasePos + 3 );
			
			// Search for the next "/../" sequence starting at the pont where we erased stuff
			findPos = erasePos;
		}
	
		return path;
	}
	
	string docURL;
	string baseURL;
	string protocol;
	string port;
	string serverName;
	string path;
	string file;
};

ParseHandler* MakeCatAnchorHandler() {
  return new CatAnchorParseHandler;
}
