/* Modified from the xmlrpc-c server example */

#include <iostream>

#include <string>
#include <vector>
#include <algorithm>

#include <xmlrpc.h>
#include <xmlrpc_abyss.h>

#include "linkdb.h"
#include "sitemap.h"
#include "tree.h"

using std::string;
using std::vector;

#define DEBUG_PRINT( x ) cerr << _x " = " << x << endl;

xmlrpc_value* sample_add( xmlrpc_env *env, xmlrpc_value *param_array, void *user_data )
{
	xmlrpc_int32 x, y, z;
	/* Parse our argument array. */
	xmlrpc_parse_value( env, param_array, "(ii)", &x, &y );
	if ( env->fault_occurred )
		return NULL;
		
	/* Add our two numbers. */
	z = x + y;
		
	/* Return our result. */
	return xmlrpc_build_value( env, "i", z );
}


/** Returns the supplied xmlrpc value unaltered. */
xmlrpc_value* stl2xmlrpc( xmlrpc_env *env, xmlrpc_value* v )
{
	return v;
}

/** Converts a C++ string to an xmlrpc value. */
xmlrpc_value* stl2xmlrpc( xmlrpc_env *env, const string& s )
{
	// Create an xmlrpc string
	xmlrpc_value* string = xmlrpc_build_value( env, "s", s.c_str() );
	if ( env->fault_occurred )
		return NULL;

	return string;
}


/** The XML-RPC Serializable interface. */
class XmlrpcSerializable
{
public:
	virtual xmlrpc_value* xmlrpcSerialize( xmlrpc_env *env ) const = 0;
};

/** Converts an XML-RPC serializable object into an xmlrpc value. */
xmlrpc_value* stl2xmlrpc( xmlrpc_env *env, const XmlrpcSerializable& obj )
{
	return obj.xmlrpcSerialize( env );
}

template <class T>
xmlrpc_value* stl2xmlrpc( xmlrpc_env *env, const vector<T>& v )
{
	return stl2xmlrpc( env, v.begin(), v.end() );
}

template <class It>
xmlrpc_value* stl2xmlrpc( xmlrpc_env *env, const It& begin, const It& end )
{
	// Create an empty array
	xmlrpc_value* array = xmlrpc_build_value( env, "()" );
	if ( env->fault_occurred )
		return NULL;
	
	for ( It i = begin; i != end; ++ i )
	{
		xmlrpc_value* value = stl2xmlrpc( env, *i );
		if ( value == NULL )
		{
			xmlrpc_DECREF( array );
			return NULL;
		}
		
		// add it to the array and release the reference
		xmlrpc_array_append_item( env, array, value );
		xmlrpc_DECREF( value );
		
		if ( env->fault_occurred )
		{
			xmlrpc_DECREF( array );
			return NULL;
		}
	}
	
	return array;
}

linkDB* db = NULL;
siteMap* sitemap = NULL;

xmlrpc_value* sitemap_exists( xmlrpc_env *env, xmlrpc_value *param_array, void *user_data )
{
	// Return an error if the database has not been initialized
	if ( db == NULL )
		return NULL;
	
	char* src;
	// Parse arguments
	xmlrpc_parse_value( env, param_array, "(s)", &src );
	if ( env->fault_occurred )
		return NULL;

	xmlrpc_bool result = db->exists( string( src ) );
	
	/* Return our result. */
	return xmlrpc_build_value( env, "b", result );
}

class URLTitlePair : public XmlrpcSerializable
{
public:
	URLTitlePair( string url, string title ) : url( url ), title( title ) {}
		
	virtual xmlrpc_value* xmlrpcSerialize( xmlrpc_env *env ) const
	{
		xmlrpc_value* retVal = xmlrpc_build_value( env, "{s:s,s:s}",
			"title", title.c_str(), "url", url.c_str() );
		if ( env->fault_occurred )
			return NULL;
		
		return retVal;
	}
	
	/** The natural sort order for URL/Title pairs is to sort by URL. */
	bool operator<( const URLTitlePair& right ) const
	{
		return url < right.url;		
	}
	
	virtual ~URLTitlePair() {}

private:
	string url;
	string title;
};


xmlrpc_value* sitemap_getLinks( xmlrpc_env *env, xmlrpc_value *param_array, void *user_data )
{
	// Return an error if the database has not been initialized
	if ( db == NULL )
		return NULL;
	
	char* src;
	// Parse arguments
	xmlrpc_parse_value( env, param_array, "(s)", &src );
	if ( env->fault_occurred )
		return NULL;
	
	vector<md5sum> result = db->getLinks( string( src ) );
	
	// Convert the links (strings) to an array of url/title structures
	vector<URLTitlePair> links;
	for ( vector<md5sum>::iterator i = result.begin(); i != result.end(); ++ i )
	{
		string url;
		string title;
		db->getURLTitlePair( *i, url, title );
		assert( url != "" );
		
		links.push_back( URLTitlePair( url, title ) );
	}
	
	// Sort the links
	std::sort( links.begin(), links.end() );	
	
	/* Return our result. */
	return stl2xmlrpc( env, links );
}


xmlrpc_value* stl2xmlrpc( xmlrpc_env *env, const TreeNode* v )
{
	string url;
	string title;
	db->getURLTitlePair( v->getValue(), url, title );
	assert( url != "" );
	
	// Recursively convert our children
	xmlrpc_value* children = stl2xmlrpc( env, v->begin(), v->end() );

	// Build the structure
	xmlrpc_value* retVal = xmlrpc_build_value( env, "{s:s,s:s,s:V}",
		"url", url.c_str(), "title", title.c_str(), "children", children );
	xmlrpc_DECREF( children );
	
	return retVal;
}

xmlrpc_value* sitemap_getSiteMap( xmlrpc_env *env, xmlrpc_value *param_array, void *user_data )
{
	// Return an error if the database has not been initialized
	if ( db == NULL )
		return NULL;
	
	char* src;
	// Parse arguments
	xmlrpc_parse_value( env, param_array, "(s)", &src );
	if ( env->fault_occurred )
		return NULL;
	
	Tree siteStructure;
	if ( ! sitemap->getSiteFromURL( md5sum( string( src ) ), siteStructure ) )
	{
		cerr << "Couldn't find site = " << src << endl;
		// Couldn't get the data: return an empty structure
		return xmlrpc_build_value( env, "{}" );
	}
	
	return stl2xmlrpc( env, &siteStructure );
}

xmlrpc_value* test_vector( xmlrpc_env *env, xmlrpc_value *param_array, void *user_data )
{
	// Parse arguments
	/*xmlrpc_parse_value( env, param_array, "()" );
	if ( env->fault_occurred )
		return NULL;*/
	
	vector<string> list;
	list.push_back( "Hello" );
	list.push_back( "World" );
	list.push_back( "How" );
	list.push_back( "are" );
	list.push_back( "you" );
	
	/* Return our result. */
	return stl2xmlrpc( env, list );
}

int main ( int argc, char **argv )
{
    if ( argc != 2 && argc != 3 )
	{
		cerr << "Usage: server <abyss.conf> [links directory]" << endl;
		return 1;
    }
	
	if ( argc == 3 )
	{
		cout << "opening " << argv[2] << endl;
		db = new linkDB( argv[2], true );
		sitemap = new siteMap( argv[2], true );
	}
	else
	{
		cout << "opening default database /tmp/bidir" << endl;
		db = new linkDB( "/tmp/bidir", true );
		sitemap = new siteMap( "/tmp/bidir", true );
	}

    xmlrpc_server_abyss_init( XMLRPC_SERVER_ABYSS_NO_FLAGS, argv[1] );
    xmlrpc_server_abyss_add_method( "sample.add", &sample_add, NULL );
	xmlrpc_server_abyss_add_method( "sitemap.exists", &sitemap_exists, NULL );
	xmlrpc_server_abyss_add_method( "sitemap.getLinks", &sitemap_getLinks, NULL );
	xmlrpc_server_abyss_add_method( "sitemap.getSiteMap", &sitemap_getSiteMap, NULL );
	xmlrpc_server_abyss_add_method( "test.vector", &test_vector, NULL );

    cout << "server: switching to background." << endl;
    xmlrpc_server_abyss_run();

    /* We never reach this point. */
    return 0;
}
