#include "sitemap.h"

//#include "basictypes.h"

#include <set>

//using std::string;

const char siteMap::DATABASE_SITES[] = "google-sites.db";
const char siteMap::DATABASE_URL_TO_SITE[] = "google-url-to-site.db";

bool siteMap::getId( const md5sum& url, md5sum& site )
{
	// Key is the url hash
	Dbt key;
	key.set_data( (void*) url.getDigest() );
	key.set_size( url.size() );
	
	// Data is the site hash
	Dbt data;

	int result = URLToSite.get( NULL, &key, &data, 0 );
	
	assert( result == 0 || result == DB_NOTFOUND );
	
	// If we managed to look up the value
	if ( result == 0 )
	{
		assert( data.get_size() == url.size() );
		
		site = md5sum( (unsigned char*) data.get_data() );
		return true;
	}
	
	return false;
}


bool siteMap::erase( const md5sum& site )
{
	// Key is the site id
	Dbt key;
	key.set_data( (void*) site.getDigest() );
	key.set_size( site.size() );
	
	// Delete the result
	int result = sites.del( NULL, &key, 0 );
	
	//cout << "erased site = " << site.toString() << endl;
	
	return ( result == 0 || result == DB_NOTFOUND );
}


vector<md5sum> siteMap::getURLs( const md5sum& site )
{
	// Get the site's structure
	Tree siteStructure;
	// The lookup failed. Return an empty vector.
	if ( ! deserializeSite( site, siteStructure ) )
		return vector<md5sum>();
		
	// Collect URLs
	URLCollector visitor;
	siteStructure.accept( visitor );
		
	// Return the urls
	vector<md5sum> retVal = visitor.getVectorURLs();
	assert( retVal.size() > 0 );
	
	return retVal;
}


bool siteMap::appendURLs( const md5sum& site, const vector<md5sum>& urls )
{
	// Don't do anything if there are no URLs to add
	if ( urls.size() == 0 )
		return true;
	
	URLCollector visitor;
	
	// Get the site's structure
	Tree siteStructure;
	if ( deserializeSite( site, siteStructure ) )
	{
		// Only go and collect URLs if we could load the data
		siteStructure.accept( visitor );
	}
	
	// Insert new URLs (we are inserting into a std::set so we don't have duplicates)
	visitor.insert( urls );
	
	// Create a new tree structure
	vector<md5sum> list = visitor.getVectorURLs();
	
	// The following assertion is not true, as urls might include duplicates
	//assert( list.size() >= urls.size() );
	assert( list.size() > 0 );
	
	siteStructure = Tree( list.front() );
	for ( vector<md5sum>::const_iterator i = list.begin() + 1; i != list.end(); ++ i )
	{
		siteStructure.appendChild( *i );
	}
	
	// Store the updated site
	if ( ! serializeSite( site, siteStructure ) )
		return false;
	
	// Now go map the urls -> site
	
	// Data is the site id
	Dbt data;
	data.set_data( (void*) site.getDigest() );
	data.set_size( site.size() );
	
	// TODO: This can probably iterate through the urls vector instead, but better safe than sorry
	for ( vector<md5sum>::const_iterator i = list.begin(); i != list.end(); ++ i )
	{
		// Key is the url id
		Dbt key;
		key.set_data( (void*) i->getDigest() );
		key.set_size( i->size() );
		
		int result = URLToSite.put( NULL, &key, &data, 0 );
		assert( result == 0 );
		
		if ( result != 0 )
			return false;
	}
	
	
	return true;
}


bool siteMap::deserializeSite( const md5sum& site, Tree& siteStructure )
{
	// Key is the site id
	Dbt key;
	key.set_data( (void*) site.getDigest() );
	key.set_size( site.size() );
	
	// Data is the serialized tree
	Dbt data;
	
	int result = sites.get( NULL, &key, &data, 0 );
	
	assert( result == 0 || result == DB_NOTFOUND );
	
	if ( result == DB_NOTFOUND )
	{
		//siteStructure = Tree();
		return false;
	}
	
	if ( result != 0 )
		return false;
	
	// Deserialize the site tree
	siteStructure = Tree( (unsigned char*)  data.get_data(), data.get_size() );
	return true;
}


bool siteMap::serializeSite( const md5sum& site, const Tree& siteStructure )
{
	// Key is the site id
	Dbt key;
	key.set_data( (void*) site.getDigest() );
	key.set_size( site.size() );

	string serialized = siteStructure.serialize();
	
	// Data is the serialized tree
	Dbt data;
	data.set_data( (void*) serialized.data() );
	data.set_size( serialized.size() );
	
	int result = sites.put( NULL, &key, &data, 0 );
	
	assert( result == 0 || result == DB_NOTFOUND );
	
	// Now go map the urls -> site
	
	// Data is the site id
	data.set_data( (void*) site.getDigest() );
	data.set_size( site.size() );
	
	URLCollector visitor;
	siteStructure.accept( visitor );
	
	// TODO: This can probably iterate through the urls vector instead, but better safe than sorry
	for ( set<md5sum>::const_iterator i = visitor.begin(); i != visitor.end(); ++ i )
	{
		// Key is the url id
		key.set_data( (void*) i->getDigest() );
		key.set_size( i->size() );
		
		int result = URLToSite.put( NULL, &key, &data, 0 );
		assert( result == 0 );
		
		if ( result != 0 )
			return false;
	}
	
	
	return true;
}

bool siteMap::accept( siteMapVisitor& visitor )
{
	Dbt key;
	Dbt data;
		
	Dbc* cursor;
	sites.cursor( NULL, &cursor, 0 );
	
	int retVal = cursor->get( &key, &data, DB_FIRST );	
	while( retVal != DB_NOTFOUND )
	{
		Tree map( (unsigned char*) data.get_data(), data.get_size() );
			
		visitor.visit( md5sum( (unsigned char *) key.get_data() ), map );
		
		retVal = cursor->get( &key, &data, DB_NEXT );
	}
	
	cursor->close();

	return true;
}

bool siteMap::accept( siteURLVisitor& visitor )
{
	Dbt key;
	Dbt data;
		
	Dbc* cursor;
	URLToSite.cursor( NULL, &cursor, 0 );
	
	int retVal = cursor->get( &key, &data, DB_FIRST );	
	while( retVal != DB_NOTFOUND )
	{			
		visitor.visit( md5sum( (unsigned char *) key.get_data() ) );
		
		retVal = cursor->get( &key, &data, DB_NEXT );
	}
	
	cursor->close();

	return true;	
}

bool siteMap::getSiteFromURL( const md5sum& url, Tree& siteStructure )
{
	// Step 1: Lookup the URL in the URLToSite database
	// Key is the url id
	Dbt key;
	key.set_data( (void*) url.getDigest() );
	key.set_size( url.size() );
	
	// Data is the site id
	Dbt data;
	
	int result = URLToSite.get( NULL, &key, &data, 0 );
	
	assert( result == 0 || result == DB_NOTFOUND );
	
	if ( result == DB_NOTFOUND )
	{
		// Couldn't find it!
		return false;
	}
	
	if ( result != 0 )
		return false;
	
	// Step 2: Get the site data and deserialize it
	return deserializeSite( md5sum( (unsigned char*) data.get_data() ), siteStructure );
}

