#ifndef _SITEMAP_H
#define _SITEMAP_H

#include <vector>
#include <string>
#include <sstream>
#include <iomanip>
#include <set>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <db_cxx.h>

#include "md5sum.h"
#include "tree.h"

class siteMapVisitor
{
	public:
		virtual void visit( const md5sum& site, const Tree& map ) = 0;
};

class siteURLVisitor
{
	public:
		virtual void visit( const md5sum& site ) = 0;
};

class siteMap
{
public:
	/** Default constructor: Creates/opens the sitemap database in the specified output path. 
	* @param specifiedOutputPath The path to create/read the databases. Defaults to "/tmp".
	* @param readOnly Specifies if the databases should be opened read only. Read only databases can have better performance. Defaults to false. */
	siteMap( const std::string& specifiedOutputPath = "", bool readOnly = false ) :
		sites( NULL, 0 ),
		URLToSite( NULL, 0 ),
		outputPath( specifiedOutputPath )
	{
		// Default to "/tmp/bidir"
		if ( outputPath.empty() )
			outputPath = "/tmp/bidir";
		
		// Verify that the path exists and is a directory
		struct stat fileInfo;
		if ( stat( outputPath.c_str(), &fileInfo ) )
		{
			// TODO: Throw a real exception here
			throw "Invalid output path (stat failed). Maybe it does not exist?";
		}
		
		// If it isn't a directory, throw an exception
		if ( ! S_ISDIR( fileInfo.st_mode ) )
		{
			//TODO: Throw a real exception here
			throw "Invalid output path (is not a directory)";
		}

		// Add a trailing slash if it was not specified
		if ( '/' != *outputPath.rbegin() )
		{
			outputPath.append( "/" );
		}
		
		u_int32_t flags = DB_CREATE;
		if ( readOnly )
			flags = DB_RDONLY;
		
		// Open the databases
		sites.open( (outputPath + DATABASE_SITES).c_str(), NULL, DB_HASH, flags, 0 );
		URLToSite.open( (outputPath + DATABASE_URL_TO_SITE).c_str(), NULL, DB_HASH, flags, 0 );
	}
	
	/** Returns true if the given url hash has a site associated with it, false otherwise. The site hash is returned in the site variable. */
	bool getId( const md5sum& url, md5sum& site );
	
	/** Returns a vector of URLs contained in the specified site. */
	std::vector<md5sum> getURLs( const md5sum& site );
	
	/** Erases the specified site from the database. */
	bool erase( const md5sum& site );
	
	/** Adds the specified URLs to the specified site, in no specific order. Duplicates will be silently ignored. */
	bool appendURLs( const md5sum& site, const vector<md5sum>& urls );
		
	/** Visit all the sites in the database. */
	bool accept( siteMapVisitor& visitor );
	
	bool accept( siteURLVisitor& visitor );
	
		
	/** Serializes the specified site from the provided tree. */
	bool serializeSite( const md5sum& site, const Tree& siteStructure ); 
	
	bool getSiteFromURL( const md5sum& url, Tree& siteStructure );
	
	
	/** Closes the open database handles. */
	~siteMap()
	{
		sites.close( 0 );
		URLToSite.close( 0 );
	}
	
	class URLCollector : public TreeVisitor
	{
		public:
			virtual void visit( const md5sum& url )
			{
				urls.insert( url );					
			}
				
			vector<md5sum> getVectorURLs()
			{
				vector<md5sum> vec;
				for ( set<md5sum>::const_iterator i = urls.begin(); i != urls.end(); ++ i )
				{
					vec.push_back( *i );
				}
						
				return vec;
			}
				
			void insert( vector<md5sum> urlList )
			{
				urls.insert( urlList.begin(), urlList.end() );
			}
				
			set<md5sum>& getURLs()
			{
				return urls;
			}
				
			set<md5sum>::const_iterator begin() const
			{
				return urls.begin();
			}
			set<md5sum>::const_iterator end() const
			{
				return urls.end();
			}
					
		private:
			set<md5sum> urls;			
	};
	
private:
	class URLCollector;
		
	/** Deserializes the specified site into the provided tree. */
	bool deserializeSite( const md5sum& site, Tree& siteStructure );

	static const char DATABASE_SITES[];
	static const char DATABASE_URL_TO_SITE[];

	Db sites;
	Db URLToSite;

	std::string outputPath;
};


#endif
