/* A basic test suite for the Trie class. It tries to exercise all the code paths. */

#include <iostream>
#include <string>

#include <fstream>

#include "tree.h"

using std::string;

class PrintingTreeVisitor : public TreeVisitor
{
	virtual void visit( const md5sum& value )
	{
		cout << value.toString() << endl;
	}
};

bool verifyTree( const TreeNode* one, const TreeNode* two )
{
	if ( one->getValue() != two->getValue() )
		return false;
	
	if ( one->size() != two->size() )
		return false;
	
	std::vector<TreeNode*>::const_iterator oneIt = one->begin();
	std::vector<TreeNode*>::const_iterator twoIt = two->begin();
	while( oneIt != one->end() )
	{
		if ( (*oneIt)->getValue() != (*twoIt)->getValue() )
		{
			return false;
		}
		
		if ( ! verifyTree( *oneIt, *twoIt ) ) return false;
		
		++ oneIt;
		++ twoIt;
	}
	
	return true;
}

void serializeTest( const Tree& t )
{
	string serialized = t.serialize();
	Tree copy( (unsigned char*) serialized.data(), serialized.size() );
	assert( verifyTree( &t, &copy ) );
}

void scopeTest( Tree t )
{
	Tree b = t;
	
	assert( verifyTree( &b, &t ) );
	
	return;
}

void scopeTest2( Tree inParam, Tree& outParam )
{
	cout << "copy: ";
	outParam = inParam;
}

void scopeTest3( Tree& one, Tree& two )
{
	Tree oneCopy( one );
	assert( verifyTree( &one, &oneCopy ) );
	Tree twoCopy( two );
	assert( verifyTree( &two, &twoCopy ) );
	
	one = two;
	assert( verifyTree( &one, &two ) );
	two = one;
	one = one;
	two = two;
}

int main()
{
	Tree test;
	
	string str[] = {
		"http://slashdot.org/",
		"http://slashdot.org/index.html",
		"http://www.scripting.com/",
		"http://www.scripting.com/index.html",
		"http://images.slashdot.org/someimage.png",
		"http://www.eng.uwaterloo.ca/~ejones/",
		"http://www.eng.uwaterloo.ca/~ejones/styles/steely.css"
		};
		
	TreeNode* a = test.appendChild( md5sum( str[0] ) );
	serializeTest( test );
	TreeNode* b = test.appendChild( md5sum( str[1] ) );
	serializeTest( test );
		
	TreeNode* c = a->appendChild( md5sum( str[2] ) );
	serializeTest( test );
		
	TreeNode* d = c->appendChild( md5sum( str[3] ) );
	serializeTest( test );
		
	b->appendChild( md5sum( str[5] ) );
	serializeTest( test );
	b->appendChild( md5sum( str[6] ) );
	serializeTest( test );
		
	d->appendChild( md5sum( str[4] ) );
	serializeTest( test );
		
	scopeTest( test );
	Tree out;
	scopeTest2( test, out );
	assert( verifyTree( &test, &out ) );
	scopeTest3( test, out );
	assert( verifyTree( &test, &out ) );
	
	// Print the contents of the tree
	cout << "Printing tree:" << endl;
	PrintingTreeVisitor print;
	test.accept( print );
	
	// Test serialization by writing the serialized version to disk, reading it
	// and reverifying the Trie contents
	cout << "Serializing... " << endl;
	string serialized = test.serialize();
	
	/*
	ofstream out( "test-tree.out", ios::binary );
	out.write( serialized.data(), serialized.length() );
	out.close();
	*/
	
	cout << "Deserializing... (" << serialized.size() << " bytes)" << endl;
	Tree copy( (unsigned char*) serialized.data(), serialized.size() );

	cout << "Printing deserialized trie:" << endl;
	copy.accept( print );

	//cout << "Verifying deserialized trie..." << endl;	
	//verifyTree( test, copy );
	assert( verifyTree( &test, &copy ) );
	
	
	// Get the file size, create a string, and slurp the whole file
	/*
	ifstream in( "bug.out", ios::binary );
	in.seekg( 0, ios::end );
	streampos fileSize = in.tellg();
	in.seekg( 0, ios::beg );
	unsigned char* buffer = new unsigned char[ fileSize ];
	in.read( buffer, fileSize );
	//if ( ! in.eof() ) die();
	
	Tree o( buffer, fileSize );
	*/
	
	return 0;
}
