#include "tree.h"

#include <stack>
#include <string>

#include <fstream>

using std::stack;
using std::string;

void Tree::accept( TreeNodeVisitor& visitor ) const
{
	stack<const TreeNode*> path;
	
	visitor.push( this );
	path.push( this );
			
	while( ! path.empty() )
	{
		const TreeNode* node = path.top();
		path.pop();
		visitor.pop( node );
		
		// Add all the children to the stack
		for ( std::vector<TreeNode*>::const_iterator i = node->begin(); i != node->end(); ++ i )
		{
			// push the node
			visitor.push( *i );
			path.push( *i );	
		}
	}
}

string Tree::serialize() const
{
	class TreeSerializer : public TreeNodeVisitor
	{
		public:
			TreeSerializer() : buffer( "" ) {}
				
			virtual void push( const TreeNode* node )
			{
				// Write out the value
				md5sum value = node->getValue();
				const unsigned char* hashBuffer = value.getDigest();
				
				for ( unsigned int i = 0; i < value.size(); ++ i )
				{
					buffer.append( 1, hashBuffer[i] );
				}
			}

			virtual void pop( const TreeNode* node )
			{
				// Write out the child count (2 bytes)
				unsigned long int childCount = node->size();
				assert( childCount < 0xffff );
				buffer.append( 1, static_cast<unsigned char>( childCount >> 8 ) );
				buffer.append( 1, static_cast<unsigned char>( childCount & 0xff ) );		
			}
			
			virtual ~TreeSerializer() {}

			string buffer;
	};
	
	TreeSerializer serialize;
	this->accept( serialize );
	
	// 18 bytes: 16 bytes for the hash and 2 bytes for the child count
	assert( serialize.buffer.size() % 18 == 0 );
	
	return serialize.buffer;
}


// The deserialize constructor
// TODO: Do some sanity checking to make sure the Tree is formatted correctly
// 	This really needs to check if the buffer is empty
Tree::Tree( const unsigned char* buffer, int length )
{
	int index = 0;
	
	assert( (unsigned) length >= this->getValue().size() );

	stack<TreeNode*> path;
	
	// Read the first md5sum
	this->setValue( md5sum( (unsigned char*) buffer ) );
	index += this->getValue().size();
	path.push( this );
	
	while( index < length )
	{
		if ( path.empty() )
		{
			ofstream out( "bug.out", ios::binary );
			out.write( buffer, length );
			out.close();
			cerr << "index = " << index;
		}
		assert( ! path.empty() );
		
		// Pop the node
		TreeNode* node = path.top();
		path.pop();
		
		// Read the next child count (2 bytes)
		int childCount = buffer[ index ++ ];
		childCount <<= 8;
		childCount |= buffer[ index ++ ];
		
		while ( childCount > 0 )
		{
			// Add the child to the current node
			TreeNode* child = node->appendChild( md5sum( (unsigned char*) buffer + index ) );
			index += this->getValue().size();
			// Push the node on the stack
			path.push( child );
			
			-- childCount;
		}
	}
	
	return;
}
