#ifndef _TREE_H
#define _TREE_H

#include <string>
#include <stack>
#include <vector>

#include "md5sum.h"

class TreeNode;

/** Visitor pattern for traversing each individual node in the Tree. This interface
	is used internally to implement some Tree routines, and is not meant to be
	used as a general base class. Use the TreeVisitor interface instead. */
class TreeNodeVisitor
{
	public:
		/** Called for each node in the preorder traversal of the Tree.
		* @param node The node that was pushed on the stack.
		* @param nodeValue The value associated with the node being visited.
		*/
		virtual void push( const TreeNode* node ) = 0;
	
		/** Called when leaving the node in the preorder traversal of the Tree. 
		* @param node The node that was popped off the stack. 
		*/
		virtual void pop( const TreeNode* node ) = 0;
};

/** Visitor pattern for traversing the Tree. */
//class TreeVisitor
//{
	//public:*
		/** Will be called once for each node visited.
		* @param character Contains the next character in the string.
		* @param isTerminal Is the string from the root to the current node complete?
		*/
		//virtual void push( char character, bool isTerminal ) = 0;
		/** Called when leaving a node. */
		//virtual void pop() = 0;
//};

/** The nodes that compose the Tree. Should not be used by external code. */
class TreeNode
{
	public:
		/** Copy constructor. */
		TreeNode( const TreeNode& rhs )
		{
			this->clone( rhs );
		}
		
		TreeNode& operator=( const TreeNode& rhs )
		{
			this->clone( rhs );
			return *this;
		}
		
		/** Creates an empty node. */
		TreeNode() {}

		/** Creates a node that stores the specified value. */
		TreeNode( const md5sum& value ) : value( value ) {}
		
		void clone( const TreeNode& rhs )
		{
			// Avoid self assignment
			if ( &rhs != this )
			{
				// Delete everything in this node
				erase();
				
				// Copy the value
				value = rhs.value;
				
				// Copy the children into new variables. This will recursively call the copy constructor.
				for ( vector<TreeNode*>::const_iterator i = rhs.childNodes.begin(); i != rhs.childNodes.end(); ++ i )
				{
					childNodes.push_back( new TreeNode( **i ) );
				}
			}
		}
		
		void erase()
		{
			while( childNodes.size() > 0 )
			{
				delete childNodes.back();
				childNodes.pop_back();
			}
		}

		const md5sum& getValue() const
		{
			return value;
		}
		
		void setValue( const md5sum& nodeValue )
		{
			value = nodeValue;
		}
		
		/** Returns a pointer to the child node with the specified index.
		* @param index The index of the child to return.
		* @return A pointer to the child node with the specified node value, or null if an error occured.
		*/
		/*
		TreeNode* getChild( std::vector<TreeNode>::size_type index )
		{
			if ( index < 0 || index > childNodes.size() )
				return NULL;
			
			return &childNodes[index];
		}
		
		const TreeNode* getChild( int index ) const
		{
			if ( index < 0 || index > childNodes.size() )
				return NULL;
			
			return &childNodes[index];
		}
		*/
	
		/** Appends a new child node to this node.
		* @param nodeValue The new node's value.
		* @return A pointer to the new child node.
		*/
		TreeNode* appendChild( const md5sum& nodeValue )
		{
			childNodes.push_back( new TreeNode( nodeValue ) );
			return childNodes.back();
		}

		/** Returns the number of child nodes contained in this node.
		* @returns The number of child nodes contained in this node.
		*/
		vector<TreeNode*>::size_type size() const
		{
			return childNodes.size();
		}
		
		/** Returns an iterator to the beginning of the child nodes. */
		std::vector<TreeNode*>::iterator begin()
		{
			return childNodes.begin();
		}
		std::vector<TreeNode*>::const_iterator begin() const
		{
			return childNodes.begin();
		}


		/** Returns an iterator to just past the end of the child nodes. */
		std::vector<TreeNode*>::iterator end()
		{
			return childNodes.end();
		}
		std::vector<TreeNode*>::const_iterator end() const
		{
			return childNodes.end();
		}
		
		~TreeNode()
		{
			erase();
		}
			
	private:
		md5sum value;
	
		/** Contains all the node value to child node associations. */
		std::vector<TreeNode*> childNodes;
};


/** Visitor pattern for interating over the values contained in the Tree. */
class TreeVisitor : public TreeNodeVisitor
{
	public:
		virtual void push( const TreeNode* node ) {}
		
		virtual void pop( const TreeNode* node )
		{
			this->visit( node->getValue() );
		}
		
		/** Called once for each node in the Tree.
		* @param value The value being visited.
		*/
		virtual void visit( const md5sum& value ) = 0;
		
		/** Default virtual destructor. Included to force it to be virtual so
			this object gets freed correctly. */
		virtual ~TreeVisitor() {}
};


class Tree : public TreeNode
{
	public:			
		Tree() : TreeNode() {}
			
		Tree( const md5sum& value ) : TreeNode( value ) {}
	
		/** Deserializes a Tree from the given character buffer.
		* @param buffer The serialized representation of the Tree.
		*/
		Tree( const unsigned char* buffer, int length );
			
		/** Traverses the Tree with the specified TreeVisitor.
		* @param visitor The instance of a TreeVisitor used to traverse the Tree.
		*/
		void accept( TreeNodeVisitor& visitor ) const;
	
		/** Serialize the Tree.
		* @return The serialized representation of the Tree.
		*/
		std::string serialize() const;
};
#endif
