#include "trie.h"

#include <stack>

using std::stack;
using std::string;

void Trie::accept( TrieNodeVisitor& visitor )
{
	stack<TrieNode*> path;
	
	// Reset the root node's iterator
	gotoFirstChild();
	visitor.push( this, '\0' );
	path.push( this );
			
	while( ! path.empty() )
	{
		TrieNode* node = path.top();
		
		// If there are more child nodes to visit,
		if ( node->hasMoreChildren() )
		{
			char nodeValue = node->nextCharacter();
			node = node->nextChildNode();
			
			// push the next node
			visitor.push( node, nodeValue );
			path.push( node );	
		}
		// If there are no more child nodes,
		else
		{
			// Pop the top off the stack
			visitor.pop();
			path.pop();
		}
	}
}

bool Trie::contains( const string& str ) const
{
	const TrieNode* nodeCursor = this;
	string::const_iterator strCursor = str.begin();
	
	while ( strCursor != str.end() )
	{
		/*
		string::const_iterator next = strCursor + 1;
		// We have reached the end of the string
		if ( next == str.end() )
		{
			// If the current node is a terminal, then this string is contained in the trie
			return nodeCursor->isTerminal();
		}
		*/
		// If the next character in the string cannot be found, the string is not in the trie
		if ( ! nodeCursor->hasChild( *strCursor ) )
		{
			return false;
		}
		
		// Compare the next character
		nodeCursor = nodeCursor->getChild( *strCursor );
		++ strCursor;
	}

	// We are at the end of the string
	return nodeCursor->isTerminal();
}

#ifndef NDEBUG
bool verifyURL( const string& url )
{
	for ( string::const_iterator i = url.begin(); i != url.end(); ++ i )
	{
		if ( static_cast<unsigned char>( *i ) & 0x80 )
		{
			return false;
		}
	}
	
	return true;	
}
#endif

void Trie::put( const string& str )
{
	// TODO: Do we need this check?
	assert( verifyURL( str ) );
	
	TrieNode* nodeCursor = this;
	string::const_iterator strCursor = str.begin();
	
	while ( strCursor != str.end() )
	{
		nodeCursor = nodeCursor->addChild( *strCursor );
		++ strCursor;
	}
	
	// We've made it to the end of the tree: terminate it
	nodeCursor->makeTerminal();
}

string Trie::serialize()
{
	class TrieSerializer : public TrieNodeVisitor
	{
		public:
			TrieSerializer() : putCount( false ), count( 0 ), buffer( "" ) {}
				
			virtual void push( TrieNode* node, char nodeValue )
			{
				// skip the root node
				if ( nodeValue == '\0' ) return;
					
				// Validation: This Trie class is made for URLs, which only
				// contain ASCII characters. The high bit should *always*
				// be unset. This is important, because we are going to store
				// data there
				assert( ! (nodeValue & 0x80) );
				
				// Make sure that the pop count gets written out
				if ( putCount )
				{
					if ( count < 256 )
					{
						buffer.append( 1, static_cast<unsigned char>( count ) );
					}
					else
					{
						assert( count  < 0xffff );
						buffer.append( 1, static_cast<unsigned char>( 0xff ) );
						buffer.append( 1, static_cast<unsigned char>( count >> 8 ) );
						buffer.append( 1, static_cast<unsigned char>( count & 0xff ) );
					}
				}
				
				putCount = false;
				count = 0;
				
				int numChildren = node->numChildNodes();
				
				if ( ! node->isTerminal() && numChildren > 0 )
				{
					buffer.append( 1, nodeValue );
					return;
				}
				buffer.append( 1, static_cast<char>( nodeValue | 0x80 ) );
				
				// if we got here, we want to record the pop count
				putCount = true;
			}

			virtual void pop()
			{
				++ count;
			}
			
			virtual ~TrieSerializer() {}
				
			bool putCount;
			int count;
			string buffer;
	};
	
	TrieSerializer serialize;
	this->accept( serialize );
	
	return serialize.buffer;
}


// The deserialize constructor
// TODO: Do some sanity checking to make sure the trie is formatted correctly
// 	This really needs to check if the buffer is empty
Trie::Trie( const unsigned char* buffer, int length )
{
	int index = 0;
	stack<TrieNode*> path;

	path.push( this );
	while( index < length )
	{
		TrieNode* node = path.top();
		
		// Read the next character
		char byte = buffer[ index ++ ];
		TrieNode* child = node->addChild( byte & ~0x80 );
		path.push( child );
		
		// If it is flagged, it is a terminal and the next byte is the pop count
		if ( byte & 0x80 )
		{
			child->makeTerminal();
			
			if ( index < length )
			{
				int popCount = buffer[ index ++ ];
				if ( popCount == 0xff )
				{
					popCount = 0;
					popCount = buffer[ index ++ ];
					popCount <<= 8;
					popCount |= buffer[ index ++ ];
				}
				while( popCount > 0 )
				{
					path.pop();
					popCount --;
				}
			}
		}		
	}
	
	return;
}
