#ifndef _TRIE_H
#define _TRIE_H

#include <string>
#include <map>
#include <stack>

class TrieNode;

/** Visitor pattern for traversing each individual node in the Trie. This interface
	is used internally to implement some Trie routines, and is not meant to be
	used as a general base class. Use the TrieVisitor interface instead. */
class TrieNodeVisitor
{
	public:
		/** Called for each node in the preorder traversal of the Trie.
		* @param node The node being visited.
		* @param nodeValue The character associated with the node being visited.
		*/
		virtual void push( TrieNode* node, char nodeValue ) = 0;
	
		/** Called when leaving the node in the preorder traversal of the Trie. */
		virtual void pop() = 0;
};

/** Visitor pattern for traversing the trie. */
//class TrieVisitor
//{
	//public:*
		/** Will be called once for each node visited.
		* @param character Contains the next character in the string.
		* @param isTerminal Is the string from the root to the current node complete?
		*/
		//virtual void push( char character, bool isTerminal ) = 0;
		/** Called when leaving a node. */
		//virtual void pop() = 0;
//};

/** The nodes that compose the Trie. Should not be used by external code. */
class TrieNode
{
	public:
		/** Creates a non-terminal node. */
		TrieNode() : terminal( false )
		{
			gotoFirstChild();
		}
		
		
		/** Flags this node as a terminal node. */
		void makeTerminal()
		{
			terminal = true;
		}
		
		/** Returns true if this is a terminal node, false otherwise.
		* @return True if this is a terminal node, false otherwise.
		*/
		bool isTerminal() const
		{
			return terminal;
		}
		
		/** Searches for a child node with the specified node value.
		* @param character The character to search for as a child of this node.
		* @return True if the node has a child node with the specified value, false otherwise.
		*/
		bool hasChild( char nodeValue ) const
		{
			return ( childNodes.find( nodeValue ) != childNodes.end() );
		}
	
		/** Returns a pointer to the child node with the specified node value.
		* @param nodeValue The value of the node to return.
		* @return A pointer to the child node with the specified node value.
		*/
		TrieNode* getChild( char nodeValue )
		{
			return &childNodes[ nodeValue ];
		}
		
		const TrieNode* getChild( char nodeValue ) const
		{
			return &(childNodes.find( nodeValue )->second);
		}
	
		/** Creates a child node with the specified value if it does not already exist.
		*	if it does exist, it returns a pointer to the existing node.
		* @param nodeValue The new node's value.
		* @return A pointer to the child node with the specified node value.
		*/
		TrieNode* addChild( char nodeValue )
		{
			if ( ! hasChild( nodeValue ) )
			{
				childNodes[ nodeValue ] = TrieNode();
			}
			return getChild( nodeValue );
		}
		
		/** Returns the number of child nodes contained in this node.
		* @returns The number of child nodes contained in this node.
		*/
		int numChildNodes()
		{
			return childNodes.size();
		}
		
		/** Returns true if there are more child nodes to iterate through.
		* @returns True if there are more child nodes to iterate through.
		*/
		bool hasMoreChildren() const
		{
			return ( cursor != childNodes.end() );
		}
		
		/** Returns the next character in the iteration. This function does not move the iteration.
		* @returns The next character in the iteration.
		*/
		char nextCharacter() const
		{
			return cursor->first;
		}
		
		/** Returns the next child node in the iteration and moves the iteration forward.
		* @returns The next child node in the iteration.
		*/
		TrieNode* nextChildNode()
		{
			// TODO: Throw an exception object of some kind
			//if ( ! hasMoreChildren() )
			//	throw "There are no more children to iterate through";
			
			std::map<char, TrieNode>::iterator current = cursor;
			++ cursor;
			
			// Reset the iterator in the node
			current->second.gotoFirstChild();
			
			return &current->second;
		}
		
		/** Resets the iteration to the beginning of the child nodes. */
		void gotoFirstChild()
		{
			cursor = childNodes.begin();
		}
			
	private:
		/** Is the path from the root to this node a valid string? */
		bool terminal;
		/** Contains all the node value to child node associations. */
		std::map<char, TrieNode> childNodes;
		std::map<char, TrieNode>::iterator cursor;
};


/** Visitor pattern for interating over the strings contained in the Trie. */
class TrieVisitor : public TrieNodeVisitor
{
	public:
		virtual void push( TrieNode* node, char nodeValue )
		{
			// The root node is given the null character
			if ( nodeValue != '\0' )
				value.append( 1, nodeValue );
			
			// an empty string can be valid
			if ( node->isTerminal() )
				this->visit( value );
		}
		
		virtual void pop()
		{
			// Remove the last character from the string
			if ( value.length() > 0 )
				value.erase( value.end() - 1 );
		}
		
		/** Called once for each string in the Trie.
		* @param str The string that is currently being visited.
		*/
		virtual void visit( const std::string& str ) = 0;
		
		/** Default virtual destructor. Included to force it to be virtual so
			this object gets freed correctly. */
		virtual ~TrieVisitor() {}
		
	private:
		string value;
};


class Trie : public TrieNode
{
	public:
		/** Creates an empty Trie. */
		Trie() {}
		
		/** Deserializes a Trie from the string.
		* @param serialized The serialized representation of the Trie.
		*/
		Trie( unsigned const char* buffer, int length );
			
		/** Traverses the Trie with the specified TrieVisitor.
		* @param visitor The instance of a TrieVisitor used to traverse the Trie.
		*/
		void accept( TrieNodeVisitor& visitor );
			
		/** Test if the given string is in the Trie.
		* @param str The string to look for.
		* @return True if the Trie contains the string, false otherwise.
		*/
		bool contains( const std::string& str ) const;
			
		/** Store the given string in the Trie.
		* @param str The string to store in the Trie.
		*/
		void put( const std::string& str );
	
		/** Serialize the Trie.
		* @return The serialized representation of the Trie.
		*/
		std::string serialize();
};
#endif
